import requests
from .helper import String
from ..errors import RequestError


class Session(object):
    '''Dynamic way to way to keep requests.Session through out whole programs.'''
    def __new__(cls,*args,**kwargs):
        if not hasattr(cls,'cache'):
            cls.cache = {}
        return super(Session,cls).__new__(cls)

    def __init__(self):
        self.session = requests.Session()


    def put_in_cache(self,url,response):
        url = url.strip()
        self.cache[url] = dict(count=1,response=response)


    def check_cache(self,url):
        """Checks if url already have a response.
        Stop from calling the request method more than once.
        Great for saving mobile data on mobile devices.
        """
        url = url.strip()
        try:
            if url in self.cache.keys():
                self.cache[url]['count'] +=1
                return self.cache[url]['response']
        except MemoryError:
            self.cache = {}
        except:
            pass 
            
    def method(self, method,url,bypass=None,**kwargs):
        """ urllib requests method """
        method = String.lower(method)
        cache = self.check_cache(url) 
        if cache and method != 'post':
            return cache

        try:
            #GET
            if method == "get":
                web = self.session.get(url,timeout=10,**kwargs)
            #POST
            if method == "post":
                web = self.session.post(url,timeout=10,**kwargs)
            #PUT
            if method == "put":
                web = self.session.put(url,timeout=10,**kwargs)
            #HEAD
            if method == "head":
                web = self.session.head(url,timeout=10)
        except requests.exceptions.InvalidURL as e:
            raise RequestError(1)
            return 
        except Exception as e:
            raise RequestError(2)
            return 
         
        status = web.status_code
        try:
            web.raise_for_status()
        except:
            raise RequestError(3)
        else:
            url = url.strip()
            self.put_in_cache(url,web)
        finally:
            return web




"""
   Custom printing function using logger. 

   Stdout can be disable on startup by setting frontend verbose flag False.
   example:: frontend.verbose = True
"""
import os
import sys 
import logging
from .. import output      

rootLogger = logging.getLogger()
rootLogger.setLevel(logging.INFO)
shell = os.environ.get('SHELL')
# use stdout if shell is None
try:
    color_text = sys.stdout.shell
except:
    class color_text():
        @classmethod
        def write(*text):
            text = text[0]
            print(text)


def handlespace(w): # for correcting joined args
    return w if any(w.endswith(x) for x in ['\n','\t','\r',' ']) else w +' '


def Print(*args):
    if args:
        output = "  ".join(args)
        if shell: # check for shell:: Python IDLE will return None
            rootLogger.info(output)
        else:
            args = list(args)
            args.extend(['\n'])
            for out in args:
                dummy_call = str(color_text.write(handlespace(out),'stdout'))
                

def Verbose(*args):
    verbose = output()
    if not verbose:
        return 
    Print(*args)





import os
import sys 
import unittest
from unittest.mock import Mock,patch
from ...pydatpiff import mixtapes as mt


class TestMixtapes(unittest.TestCase):
    def testCategory(self):
        """testing Mixtapes"""
        # testing category
        results = mt.Mixtapes('hot').artists
        self.assertIsNotNone(results)

    @patch.object(mt.Mixtapes,'_Start')
    def test_start_function(self,start):
        start(start='lil wayne')
        mix = mt.Mixtapes
        start.assert_called_once()
        start.assert_called_once_with(start='lil wayne')
     
    @patch.object(mt.Mixtapes,'_Start')
    def test_start_function(self,start):
        start(start='lil wayne')
        mix = mt.Mixtapes
        start.assert_called_once()
        start.assert_called_once_with(start='lil wayne')
        
     
        



import re
from .webhandler import Html
from ..urls import Urls
from ..utils.request import Session
from ..errors import AlbumError

#TODO::Change album to just get start url response embed url response

class Album():
    def __init__(self,link):
        self.link = ''.join((Urls.url['album'] ,link))
        self._session = Session()


    @property
    def name(self):
        text = self.embed_response.text
        return re.search(r'title">(.*[\w\s]*)\</div',text).group(1)


    @property
    def ID(self):
        """Album ID Number  """
        return Html.get_end_digits(self.link)


    @property
    def embed_url(self):
        """Creates a url link to Datpiff album's media player."""

        ref_number = self.ID
        embed_link = "".join(('https://embeds.datpiff.com/mixtape/', 
                                str(ref_number),
                                '?trackid=1&platform=desktop'))
        return embed_link


    @property
    def embed_response(self):
        """Return the response content of the embedUrl. SEE: embedUrl"""
        try:
            response = self._session.method('GET', self.embed_url)
        except:
            raise AlbumError(2)
        return response


    
class Mp3():
    def __init__(self,response):
        self.response = response


    def __len__(self):
        if self.songs:
            return len(self.songs)


    @property
    def content(self):
        """Requests response from album url link See: Album.embed_url."""
        try:
            return self.response.text
        except:
            #Not a requests object
            raise Mp3Error(2)


    @property
    def song_duration(self):
        """Duration of songs"""
        return Html.get_duration_from(self.content)


    @property
    def songs(self):
        """Songs from mixtape album."""
        return Html.find_song_names(self.content)


    @property
    def prefix_tracks(self):
        """
        Track prefix song  
        Return url encoded song index joined with song name 
        Ex: 02) - Off the Wall.mp3' -->  02)%20-%20Off%20the%20Wall.mp3
        """
        return Html.find_name_of_mp3(self.content)


    @property
    def media_id(self):
        """
        Media Album reference ID number 
        
        Ex: 6/m1393dba
        """
        try:
            return Html.toId(self.content)
        except:
            Mp3Error(1)

    
    
    @property
    def mp3Urls(self):
        mp3 = []
        prefix = 'https://hw-mp3.datpiff.com/mixtapes/'
        for track in self.prefix_tracks:
            endpoint = '{}{}'.format(self.media_id,track)
            yield ''.join((prefix,endpoint))

            



import re

class String():
  @classmethod
  def lower(cls,string):
    ''' return lowered and strip string'''
    return string.lower().strip()

  @classmethod
  def title(cls,string):
    return string.lower().strip().title()

  @classmethod
  def filter(cls,string,pattern,pattern_two):
    return re.sub(pattern,pattern_two,string)





  



class Urls:
    url = {
        'base'  :'https://www.datpiff.com/',
        'album' :'http://www.datpiff.com',
        'search':'https://www.datpiff.com/mixtapes-search',
        }

    category = {
                "hot":"http://www.datpiff.com/mixtapes/hot",
                "new":"http://www.datpiff.com/mixtapes",
                "top":"http://www.datpiff.com/mixtapes-top",
                "celebrated":"http://www.datpiff.com/mixtapes/celebrated",
                "popular":"http://www.datpiff.com/mixtapes-popular.php",
                "exclusive":"http://www.datpiff.com/mixtapes-exclusive",
                "most download":"http://www.datpiff.com/mixtapes-popular.php?sort=downloads",
                "most listen":"http://www.datpiff.com/mixtapes-popular.php?filter=month&sort=listens",
                "most favorite":"http://www.datpiff.com/mixtapes-popular.php?sort=favorites",
                "highest rating":"http://www.datpiff.com/mixtapes-popular.php?filter=month&sort=rating"
                }

    @staticmethod
    def payload(artist):
        data = {'submit':'MTAxNTUuNzcxNTI5NDEyMzY0MTgwNzEx',
                        'criteria':artist
               }
        return data




import os
import sys
import re
from functools import wraps
from .urls import Urls
from .utils.request import Session
from .errors import MixtapesError
from .frontend.display import Print,Verbose
from .backend.config import User,Datatype
from .backend.webhandler import Html


class Mixtapes(object):
    def __new__(cls, *args, **kwargs):
        cls._category = Urls.category
        return super(Mixtapes, cls).__new__(cls)


    def __init__(self, category=None,search=None,*args,**kwargs):
        '''
        Mixtapes Initialize 

        @@params: category:  -- see Mixtapes.category
        '''
        super(Mixtapes,self).__init__(*args,**kwargs)
        self._session = Session()
        self._Start(category,search)
        self._setup()


    def __str__(self):
        category = "'new','hot','top','celebrated'"
        return "%s(category)  argument: category --> %s"%(self.__class__.__name__, category)


    def __repr__(self):
        return "%s('hot')"%self.__class__.__name__


    def __len__(self):
        if hasattr(self,'_artists'):
            return len(self._artists)
        else:
            return 0


    def search(self,artist):
        """search for an artist mixtapes."""
        artist = str(artist).strip()
        if not artist: 
            return 

        Verbose('\nSearching for %s mixtapes ...'%artist)
        url = Urls.url['search']
        return  self._session.method('POST',url,data=Urls.payload(artist))


    def _Start(self,category='hot', search=None):
        '''
        Starts the web page from category selected by user
        (see Mixtapes.__init__)
        '''
        if search: # Search for an artist 
            body = self.search(search)
            if not body or body is None: # on failure return reponse from a category 
                return self._Start('hot')
        else: # Select from category instead of searching 
            select = User.choice_is_str
            choice = select(category,self._category) or select('hot',self._category)
            body = self._session.method('GET',choice)
        self._responses = body
        return body

   
    def _setup(self):
        """Initial variable and set attributes on page load up."""
        self.artists = '<div class\="artist">(.*[.\w\s]*)</div>'
        self.mixtapes  = '"\stitle\="listen to ([^"]*)">[\r\n\t\s]?.*img'
        self.links   = 'title"><a href\=\"(.*[\w\s]*\.html)"'
        self.views   = '<div class\="text">Listens: <span>([\d,]*)</span>'
        Verbose('Found %s Mixtapes'%len(self))


    @property
    def category(self):
        """All category of Mixtapes that users can choose from."""
        return list(self._category)


    def _searchTree(f):
        '''
        Wrapper function that parse and filter all requests response content.
        After parsing and filtering the responses text, it then creates a 
        dunder variable from the parent function name. 
        
        @@params: f  (wrapper function) 
         #----------------------------------------------------------
         example: "function()" will create an attribute "_function"
         #----------------------------------------------------------
        '''
        @wraps(f) 
        def inner(self, *args,**kwargs):
            response_text  = self._responses.text
            name = f.__name__
            path = f(self,*args,**kwargs)
            pattern = re.compile(path)
            data = list(Html.remove_ampersands(pat.group(1))[0]\
                    for pat in pattern.finditer(response_text)\
                    if pat is not None)

            if hasattr(self,'_artists'):
                # we map all attributes length to _artists length 
                data = data[:len(self._artists)]
            dunder = '_'+name
            setattr(self,dunder,data)
            return data
        return inner


    @property
    def artists(self):
        ''' return all Mixtapes artists name'''
        if hasattr(self,'_artists'):
            return self._artists

    @artists.setter
    @_searchTree
    def artists(self,path):
        return path


    @property
    def mixtapes(self):
        if hasattr(self,'_mixtapes'):
            return self._mixtapes

    @mixtapes.setter 
    @_searchTree
    def mixtapes(self,path):
        ''' return all the mixtapes titles from web page'''
        return path

    @property
    def links(self):
        """Return all of the Mixtapes url links"""
        if hasattr(self,'_links'):
            return self._links

    @links.setter
    @_searchTree
    def links(self,path):
        ''' return all the album links from web page'''
        return path
    

    @property
    def views(self):
        """Return the view count of each mixtapes"""
        if hasattr(self,'_views'):
            return self._views

    @views.setter
    @_searchTree
    def views(self,path):
        return path


    @property
    def display(self):
        ''' Prettify all Mixtapes information  
            and display it to screen 
        '''
        links = self._links
        data = zip(self._artists, self._mixtapes, links, self._views)
        for count,(a, t, l, v) in list(enumerate(data,start=1)):
            Print("# %s\nArtist: %s\nAlbum: %s\nLinks: %s\nViews: %s\n%s"
                    % (count, a, t, l[1:], v, "-"*40))


    def _select(self,select):
        '''
        User select function that queue an ablum to the media player
        
        @@params:: select - (int) user selection by indexing an artist name or album name
                            (str)
        '''
        try:
            # Return the user select by either integer or str
            # we map the integer to artists and str to mixtapes 
            return User.selection(select,self.artists,self.mixtapes)

        except MixtapesError as e:
            Print(e)
            raise MixtapesError(1)




import re
import vlc 
from .frontend.display import Print
from time import sleep
from .errors import PlayerError


class Player(object):
    """Media player controller""" 

    def __init__(self):
        try:
            self._vlc = vlc.Instance('-q')
            self._player = self._vlc.media_player_new()
        except Exception as e:
            extended_msg = 'Please check if your device supports VLC'
            raise PlayerError(1,extended_msg)

        self._track_set = False
        self.song = None
   

    def _seeker(self,pos=10,rew=True):
        if self._state == 'No Media':
            return 
        if rew: 
            to_postion = self.player.get_time() - (pos * 1000)
        else:
            to_postion = self.player.get_time() + (pos * 1000)
        self.player.set_time(to_postion)


    @property
    def player(self):
        return self._player
    

    @property
    def _state(self):
        """Current state of the song being played"""
        state = re.match(r'[\w.]*\.(\w*)',str(self.player.get_state())).group(1)
        state = 'No Media' if state =='NothingSpecial' else state
        return state

       

    def setTrack(self,song,media=None):
        if  media:
            self.song = song
            self.player.set_mrl(media)
            self._track_set = True
        else:
            Print('No media to play')
        

    property
    def _format_time(self,pos):
        """Format current song time to clock format"""
        mins = int(pos/60000)
        secs = int((pos%60000)/1000)
        return mins,secs

    

    @property
    def info(self):
        """Returns feedback for media song being played"""
        if self._state == 'No Media':
            return 'No media' 
        c_min,c_sec = self._format_time(self.player.get_time())
        c_sec = c_sec if len(str(c_sec)) >1 else str(c_sec).zfill(2) 

        l_min,l_sec = self._format_time(self.player.get_length())
        l_sec = l_sec if len(str(l_sec)) >1 else str(l_sec).zfill(2) 
        
        Print('TRACK:',self.song)
        Print('MODE:',self._state)
        pos = 'POSITION: {0}:{1} - {2}:{3}'.format(c_min,c_sec,l_min,l_sec)
        Print(pos)
             

    @property
    def _current_volume(self):
        """ Current media player volume"""
        return self.player.audio_get_volume()


    def _set_volume(self,vol=5,way='down'):
        """Turn the media volume up or down"""
        if isinstance(vol,(int,str)):
            if not str(vol).isnumeric():
                return 

            vol = int(vol)
            min_vol = 0
            max_vol = 100
            try:
                current_volume = int(self._current_volume)
                if way == 'down':
                    if current_volume - vol < min_vol:
                        vol = min_vol 
                    else:
                        vol = current_volume - vol
                elif way == 'up':
                    if current_volume + vol > max_vol:
                        vol = max_vol
                    else:
                        vol = current_volume + vol
                elif way == 'exact':
                    vol = 0 if vol < min_vol else vol 
                    vol = 100 if vol > max_vol else vol
                Print('volume: %s'%vol)
            except:
                pass
        self.player.audio_set_volume(vol)


    def volumeUp(self,vol=5):
        """Turn the media volume up"""
        self._set_volume(vol,way='up')
     

    def volumeDown(self,vol=5):
        """Turn the media volume down"""
        self._set_volume(vol,way='down')


    def volume(self,vol=None):
        """Set the volume to exact number"""
        if vol is None:
            return
        self._set_volume(vol,way='exact')
 
   
    @property
    def play(self):
        """ Play media song"""
        playing = True if self.player.is_playing() >=1 else False

        if playing:
            # call play while playing ,them pause 
            self.player.pause()
        elif self._state == 'Paused':
            self.player.pause()

        elif self._track_set:
            self.player.play()
            self._track_set = False
        return


   
    @property
    def pause(self):
        """Pause the media song"""
        self.player.pause()


    def rewind(self,pos=10):
        """
        Rewind the media song
             vlc time is in milliseconds
             @params: pos:: time(second) to rewind media. default:10(sec)
        """
        self._seeker(pos,True)


    def ffwd(self,pos=10):
        """Fast forward media 
             vlc time is in milliseconds
             @params: pos:: time(second) to rewind media. default:10(sec)
        """
        self._seeker(pos,False)


       
    @property
    def stop(self):
        self.player.stop()


    def close(self):
        self.stop

      
     
